#include <stdio.h>
#include "fwHelpers.h"

/*******************************************
 * Kernel to improve the shortest paths by considering the kth vertex
 * as an intermediate.
 *
 * arg:  k = the index of the vertex we are using as intermediate
 * arg:  devArray = the array containing the matrix
 * arg:  N  = the number of vertices in the graph
 * arg:  Na = the width of the matrix as stored on the device's memory.
 *
 * The graph is stored as an N x Na matrix, with the (i, j) matrix entry
 * being stored as devArray[i*Na + j]
 *
 *******************************************/
__global__ void useIntermediate(int k, int devArray[], int Na, int N){
  int col = blockIdx.x * blockDim.x + threadIdx.x; /* This thread's matrix column */
  if(col >= N)
    return;

  int arrayIndex = Na * blockIdx.y + col;
  __shared__ int trkc;	       /* this row, kth column */
  
  // Improve by using the intermediate k, if we can
  if(threadIdx.x == 0)
    trkc = devArray[Na * blockIdx.y + k];
  __syncthreads();

  if(trkc == INT_MAX)		   /* infinity */
    return;

  int tckr = devArray[k*Na + col]; /* this column, kth row */
  if(tckr == INT_MAX)		   /* infinity */
    return;

  int betterMaybe = trkc + tckr;
  if(betterMaybe < devArray[arrayIndex])
    devArray[arrayIndex] =  betterMaybe;
}



/******************************************************
 * main
 *
 * We read in the graph from a graph file, in this format:
 * First we read in N, the number of vertices
 * The vertices are assumed to be numbered 1..N
 * Then we read in triples of the form s d w
 *   where s is the source vertex, d the destination, and w the weight
 *
 * The "best weight so far" matrix stores rows contiguously, with
 * bwsf(i, j) being the best weight FROM i TO j
 *
 * Thread blocks comprise rows of the matrix, so that we can
 * take advantage of global memory access grouping
 ********************************************************/
int main(int argc, char* argv[]){
  int N = 0;			/* Number of vertices                      */
  int Na = 0;			/* Width of matrix to encourage coalescing */

  int* graph = readGraph(N, Na, argv[1]); /* from fwHelpers.cpp            */
  printf("Kernel: Just read %s with %d vertices, Na = %d\n", argv[1], N, Na);

  // Copy the array to newly-allocated global memory
  int* devArray;
  cudaError_t err = cudaMalloc(&devArray, Na*N*sizeof(int));
  printf("Malloc device rules: %s\n",cudaGetErrorString(err));
  err = cudaMemcpy(devArray, graph, Na*N*sizeof(int), cudaMemcpyHostToDevice);
  printf("Pre-kernel copy memory onto device: %s\n",cudaGetErrorString(err));

  // Set up and run the kernels
  int threadsPerBlock = 256;
  dim3 blocksPerGrid((Na + threadsPerBlock - 1) / threadsPerBlock, N);

  // The kth run through this loop considers whether we might do better using
  // the kth vertex as an intermediate
  for(int k = 0; k < N; k++){
    useIntermediate <<< blocksPerGrid, threadsPerBlock >>> (k, devArray, Na, N);
    err = cudaThreadSynchronize();

    // Uncomment the following line when debugging the kernel
    // printf("Kernel: using %d as intermediate: error = %s\n", k, cudaGetErrorString(err));

    // Uncomment the following two lines to print intermediate results
    // err = cudaMemcpy(graph, devArray, Na*N*sizeof(int), cudaMemcpyDeviceToHost);
    // printArray(Na, graph);
  }

  err = cudaMemcpy(graph, devArray, Na*N*sizeof(int), cudaMemcpyDeviceToHost);
  printf("Post-kernel copy memory off of device: %s\n",cudaGetErrorString(err));

  printArray(Na, graph);
  free(graph);
  cudaFree(devArray);
}



